/*
  Version 0.4 - April 26 2019
*/

#define ledR D5
#define ledG D6
#define ledB D7

#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsClient.h> //  https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries
#include <ArduinoJson.h> // https://github.com/kakopappa/sinric/wiki/How-to-add-dependency-libraries (use the correct version)
#include <StreamString.h>

ESP8266WiFiMulti WiFiMulti;
WebSocketsClient webSocket;
WiFiClient client;

#define MyApiKey "cxxxxxxxxxxxxxxxxxxxxed8" // TODO: Change to your sinric API Key. Your API Key is displayed on sinric.com dashboard
#define MySSID "Lxxxxxxxxx" // TODO: Change to your Wifi network SSID
#define MyWifiPassword "1xxxxxx3" // TODO: Change to your Wifi network password
#define DeviceID "5dcxxxxxxxxxxxxxxxxxxxxxxd"

#define HEARTBEAT_INTERVAL 300000 // 5 Minutes 

uint64_t heartbeatTimestamp = 0;
bool isConnected = false;
bool luce = false;
bool oldState = false;

float hue = 0.0;
float sat = 0.0;
float bri = 0.0;
int brigh = 100;

// deviceId is the ID assgined to your smart-home-device in sinric.com dashboard. Copy it from dashboard and paste it here

void turnOn(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn on device id: ");
    Serial.println(deviceId);
    luce = true;
  }
  else {
    Serial.print("Turn on for unknown device id: ");
    Serial.println(deviceId);
  }
}

void turnOff(String deviceId) {
  if (deviceId == DeviceID) // Device ID of first device
  {
    Serial.print("Turn off Device ID: ");
    Serial.println(deviceId);
    luce = false;
  }
  else {
    Serial.print("Turn off for unknown device id: ");
    Serial.println(deviceId);
  }
}

void webSocketEvent(WStype_t type, uint8_t * payload, size_t length) {
  switch (type) {
    case WStype_DISCONNECTED:
      isConnected = false;
      Serial.printf("[WSc] Webservice disconnected from sinric.com!\n");
      break;
    case WStype_CONNECTED: {
        isConnected = true;
        Serial.printf("[WSc] Service connected to sinric.com at url: %s\n", payload);
        Serial.printf("Waiting for commands from sinric.com ...\n");
      }
      break;
    case WStype_TEXT: {
        Serial.printf("[WSc] get text: %s\n", payload);
        // Example payloads

        // For Switch or Light device types
        // {"deviceId": xxxx, "action": "setPowerState", value: "ON"} // https://developer.amazon.com/docs/device-apis/alexa-powercontroller.html

        // For Light device type
        // Look at the light example in github

        DynamicJsonBuffer jsonBuffer;
        JsonObject& json = jsonBuffer.parseObject((char*)payload);
        String deviceId = json ["deviceId"];
        String action = json ["action"];
        if (deviceId == DeviceID)
        {
          if (action == "setPowerState") { // Switch or Light
            String value = json ["value"];
            if (value == "ON") {
              turnOn(deviceId);
            } else {
              turnOff(deviceId);
            }
          }
          else if (action == "SetColor") {
            // Alexa, set the device name to red
            // get text: {"deviceId":"xxxx","action":"SetColor","value":{"hue":0,"saturation":1,"brightness":1}}
            String Ahue = json ["value"]["hue"];
            String Asaturation = json ["value"]["saturation"];
            String Abrightness = json ["value"]["brightness"];

            hue = Ahue.toInt();
            sat = Asaturation.toDouble();
            bri = Abrightness.toDouble();

            Serial.println("[WSc] hue: " + Ahue);
            Serial.println("[WSc] saturation: " + Asaturation);
            Serial.println("[WSc] brightness: " + Abrightness);
            luce = true;
          }
          else if (action == "SetColorTemperature") {
            String Ahue = json ["value"];
            String Asaturation = "1";
            String Abrightness = "1";

            hue = Ahue.toDouble() / 55.0;
            sat = Asaturation.toDouble();
            bri = Abrightness.toDouble();
            luce = true;
          }
          else if (action == "SetBrightness") {
            String Abrightness = json ["value"];;

            brigh = Abrightness.toInt();
            luce = true;
          }
          else if (action == "test") {
            Serial.println("[WSc] received test command from sinric.com");
          }
        } else {
          Serial.println("[WSc] Nope!");
        }
      }
      break;
    case WStype_BIN:
      Serial.printf("[WSc] get binary length: %u\n", length);
      break;
  }
}

void setup() {
  pinMode(ledR, OUTPUT);
  pinMode(ledG, OUTPUT);
  pinMode(ledB, OUTPUT);

  Serial.begin(115200);

  WiFiMulti.addAP(MySSID, MyWifiPassword);
  Serial.println();
  Serial.print("Connecting to Wifi: ");
  Serial.println(MySSID);

  // Waiting for Wifi connect
  while (WiFiMulti.run() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  if (WiFiMulti.run() == WL_CONNECTED) {
    Serial.println("");
    Serial.print("WiFi connected. ");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }

  // server address, port and URL
  webSocket.begin("iot.sinric.com", 80, "/");

  // event handler
  webSocket.onEvent(webSocketEvent);
  webSocket.setAuthorization("apikey", MyApiKey);

  // try again every 5000ms if connection has failed
  webSocket.setReconnectInterval(5000);   // If you see 'class WebSocketsClient' has no member named 'setReconnectInterval' error update arduinoWebSockets
}

void loop() {
  webSocket.loop();

  if (isConnected) {
    uint64_t now = millis();

    // Send heartbeat in order to avoid disconnections during ISP resetting IPs over night. Thanks @MacSass
    if ((now - heartbeatTimestamp) > HEARTBEAT_INTERVAL) {
      heartbeatTimestamp = now;
      webSocket.sendTXT("H");
      if (luce) {
        setPowerStateOnServer(DeviceID, "ON");
      } else {
        setPowerStateOnServer(DeviceID, "OFF");
      }
    }
  }
  if (luce) {
    setLedColorHSV(hue, sat, bri);
  } else {
    digitalWrite(ledR, LOW);
    digitalWrite(ledG, LOW);
    digitalWrite(ledB, LOW);
  }
  if (luce != oldState) {
    if (luce) {
      setPowerStateOnServer(DeviceID, "ON");
    } else {
      setPowerStateOnServer(DeviceID, "OFF");
    }
    oldState = luce;
  }
}

void setLedColorHSV(int h, double s, double v) {
  //this is the algorithm to convert from RGB to HSV
  double r = 0;
  double g = 0;
  double b = 0;

  double hf = h / 60.0;

  int i = (int)floor(h / 60.0);
  double f = h / 60.0 - i;
  double pv = v * (1 - s);
  double qv = v * (1 - s * f);
  double tv = v * (1 - s * (1 - f));

  switch (i)
  {
    case 0: //rojo dominante
      r = v;
      g = tv;
      b = pv;
      break;
    case 1: //verde
      r = qv;
      g = v;
      b = pv;
      break;
    case 2:
      r = pv;
      g = v;
      b = tv;
      break;
    case 3: //azul
      r = pv;
      g = qv;
      b = v;
      break;
    case 4:
      r = tv;
      g = pv;
      b = v;
      break;
    case 5: //rojo
      r = v;
      g = pv;
      b = qv;
      break;
  }

  //set each component to a integer value between 0 and 255
  int red = constrain((int)255 * r, 0, 255);
  int green = constrain((int)255 * g, 0, 255);
  int blue = constrain((int)255 * b, 0, 255);

  setLedColor(red, green, blue);
}

//Sets the current color for the RGB LED
void setLedColor(int red, int green, int blue) {
  //Note that we are reducing 1/4 the intensity for the green and blue components because
  //  the red one is too dim on my LED. You may want to adjust that.
  analogWrite(ledR, (((red * brigh) / 100) * 4));
  analogWrite(ledG, (((green * brigh) / 100) * 4));
  analogWrite(ledB, (((blue * brigh) / 100) * 4));
}

void setPowerStateOnServer(String deviceId, String value) {
  DynamicJsonBuffer jsonBuffer;
  JsonObject& root = jsonBuffer.createObject();
  root["deviceId"] = deviceId;
  root["action"] = "setPowerState";
  root["value"] = value;
  StreamString databuf;
  root.printTo(databuf);

  webSocket.sendTXT(databuf);
}
